package com.chrismweb.box2dforflashgames
{
	import flash.display.Sprite;
	import flash.events.Event;
	import flash.display.Sprite
	import Box2D.Dynamics.*;
	import Box2D.Collision.*;
	import Box2D.Collision.Shapes.*;
	import Box2D.Common.Math.*;
	import flash.events.Event;
	import flash.events.MouseEvent;
	
	import flash.text.TextField;
	import flash.text.TextFormat;
	import flash.text.TextFormatAlign;
	
	
	/**
	 * ...
	 * #Chp4.1 - Falling "apples" - showing multiple ways to move 3 identical objects
	 * @book by Emanuele Feronato
	 * http://www.emanueleferonato.com/
	 * Get it now @ http://www.packtpub.com/box2d-for-flash-games/book
	 */
	public class Main extends Sprite 
	{
		private var world:b2World;
		private var worldScale:Number = 30;
		private var sphereVector:Vector.<b2Body>;
		
		public function Main():void 
		{
			if (stage) 
				Init();
			else 
				addEventListener(Event.ADDED_TO_STAGE, Init);
		}
		
		private function Init(e:Event = null):void 
		{
			removeEventListener(Event.ADDED_TO_STAGE, Init);
			// entry point

			var gravity:b2Vec2 = new b2Vec2(0, 10);
			var sleep:Boolean = true;
			world = new b2World(gravity, sleep); //need to make sure you removed 'var', otherwise it will only initialize it as a local variable
			
			Setup();
			
			//for enabling debug draw
			var debugSprite:Sprite = new Sprite();
			addChild(debugSprite);
			
			var debugDraw:b2DebugDraw = new b2DebugDraw();
			debugDraw.SetSprite(debugSprite);
			debugDraw.SetDrawScale(worldScale);
			debugDraw.SetFlags(b2DebugDraw.e_shapeBit);
			debugDraw.SetFillAlpha(0.5);
			world.SetDebugDraw(debugDraw);
			
			
			
			//stuff needed to reset example
			var reset_txt:TextField = new TextField();			
			var format:TextFormat = new TextFormat();
			format.color = 0xFFFFFFFF;
			format.align = TextFormatAlign.CENTER;
			format.size = 20;
			format.font = "Arial";
			
			reset_txt.width = 100;
			reset_txt.height = 30;
			reset_txt.selectable = false;
			reset_txt.text = "Reset";
			reset_txt.textColor = 0xFFFFFFFF;
			reset_txt.backgroundColor = 0xFF0D3451;
			reset_txt.opaqueBackground = reset_txt.backgroundColor;
			reset_txt.border = true;
			reset_txt.borderColor = 0xFFFFFFFF;
			reset_txt.defaultTextFormat = format;
			reset_txt.setTextFormat(format);
			addChild(reset_txt);
			reset_txt.addEventListener(MouseEvent.CLICK, function(e:MouseEvent):void { ResetWorld() } );
			
			addEventListener(Event.ENTER_FRAME,UpdateWorld);
		}
		
		private function ResetWorld():void
		{
			trace("Resetting world");
			for (var body:b2Body=world.GetBodyList();body!=null; body=body.GetNext())
				world.DestroyBody(body);
				
			Setup();
		}
		private function Setup():void
		{
			Floor();
			sphereVector=new Vector.<b2Body>();
			for (var i:int = 0; i < 3; i++) 
			{
				sphereVector.push(Sphere(170+i*150,410,40));
			}
			
			
			//apply forces, impulses, linear velocity:
			var force:b2Vec2 = new b2Vec2(0, -15);
			var forceByMass:b2Vec2 = force.Copy();
			forceByMass.Multiply(sphereVector[1].GetMass());
			
			var forceByMassByTime:b2Vec2 = forceByMass.Copy();
			forceByMassByTime.Multiply(30); //because the timestep for box2d is currently set to 30
			
			var sphereCenter:b2Vec2 = sphereVector[0].GetWorldCenter();
			
			
			//sphereVector[0].ApplyForce(force, sphereCenter);
			sphereVector[0].ApplyForce(forceByMassByTime,sphereCenter);
			
			
			sphereCenter=sphereVector[1].GetWorldCenter();
			//sphereVector[1].ApplyImpulse(force, sphereCenter);
			
			sphereVector[1].ApplyImpulse(forceByMass,sphereCenter);
			
			sphereVector[2].SetLinearVelocity(force);
		}
		
		private function Sphere(pX:int, pY:int, r:Number):b2Body 
		{
			var bodyDef:b2BodyDef=new b2BodyDef();
			bodyDef.position.Set(PixToMeter(pX),PixToMeter(pY));
			bodyDef.type = b2Body.b2_dynamicBody;
			bodyDef.userData = pY;
			
			var circleShape:b2CircleShape
			circleShape = new b2CircleShape(PixToMeter(r));
			
			var fixtureDef:b2FixtureDef=new b2FixtureDef();
			fixtureDef.shape=circleShape;
			fixtureDef.density=2;
			fixtureDef.restitution=0.4;
			fixtureDef.friction = 0.5;
			
			var theSphere:b2Body=world.CreateBody(bodyDef);
			theSphere.CreateFixture(fixtureDef);
			return theSphere;
		}
		private function Floor():void 
		{
			var bodyDef:b2BodyDef=new b2BodyDef();
			bodyDef.position.Set(PixToMeter(320), PixToMeter(465));
			
			var polygonShape:b2PolygonShape=new b2PolygonShape();
			polygonShape.SetAsBox(PixToMeter(320), PixToMeter(15));
			
			var fixtureDef:b2FixtureDef=new b2FixtureDef();
			fixtureDef.shape=polygonShape;
			fixtureDef.restitution=0.4;
			fixtureDef.friction = 0.5;
			
			var theFloor:b2Body=world.CreateBody(bodyDef);
			theFloor.CreateFixture(fixtureDef);
		}
		
		private function PixToMeter(pixels:Number):Number
		{
			return pixels / worldScale;
		}
		private function UpdateWorld(e:Event):void 
		{
			var timeStep:Number = 1 / 30;
			var velIterations:int=10;
			var posIterations:int=10;
			world.Step(timeStep, velIterations, posIterations);
			world.ClearForces();
			world.DrawDebugData();
			
			var maxHeight:Number;
			var currHeight:Number;
			var outHeight:Number;
			
			for (var i:int = 0; i < 3; i++) 
			{
				maxHeight=sphereVector[i].GetUserData();
				currHeight=sphereVector[i].GetPosition().y*worldScale;
				maxHeight=Math.min(maxHeight,currHeight);
				sphereVector[i].SetUserData(maxHeight);
				outHeight=sphereVector[i].GetUserData();
				trace("Sphere "+i+":"+Math.round(outHeight));
			}
			trace("---------------");
		}
	}
	
}